import papis
import sys
import os
import shutil
import papis.api
import papis.utils


class Command(papis.commands.Command):
    def init(self):

        self.parser = self.get_subparsers().add_parser(
            "rm",
            help="Delete entry"
        )

        self.add_search_argument()

        self.parser.add_argument(
            "--file",
            help="Remove files from a document instead of the whole folder",
            default=False,
            action="store_true"
        )

        self.parser.add_argument(
            "-f", "--force",
            help="Do not confirm removal",
            default=False,
            action="store_true"
        )

    def main(self):
        documents = papis.api.get_documents_in_lib(
            self.get_args().lib,
            self.get_args().search
        )
        document = self.pick(documents)
        if not document: return 0
        if self.get_args().file:
            filepath = papis.api.pick(
                document.get_files()
            )
            if not filepath: return 0
            if not self.args.force:
                if not papis.utils.confirm("Are you sure?"):
                    return 0
            print("Removing %s..." % filepath)
            document.rm_file(filepath)
            document.save()
        else:
            folder = document.get_main_folder()
            if not self.args.force:
                if not papis.utils.confirm("Are you sure?"):
                    return 0
            print("Removing %s..." % folder)
            shutil.rmtree(folder)
            papis.api.clear_lib_cache()



import papis
import os
import sys
import shutil
import papis.utils
import papis.bibtex
import papis.downloaders.utils
import papis.api
import papis.document
import papis.isbn


class Command(papis.commands.Command):
    def init(self):

        self.parser = self.get_subparsers().add_parser(
            "update",
            help="Update a document from a given library"
        )

        self.add_search_argument()

        self.parser.add_argument(
            "-i",
            "--interactive",
            help="Interactively update",
            default=False,
            action="store_true"
        )

        self.parser.add_argument(
            "-f",
            "--force",
            help="Force update, overwrite conflicting information",
            default=False,
            action="store_true"
        )

        self.parser.add_argument(
            "-d",
            "--document",
            help="Overwrite an existing document",
            default=None,
            action="store"
        )

        self.parser.add_argument(
            "--from-isbnplus",
            help="Update info from isbnplus.org",
            action="store"
        )

        self.parser.add_argument(
            "--from-yaml",
            help="Update info from yaml file",
            action="store"
        )

        self.parser.add_argument(
            "--from-bibtex",
            help="Update info from bibtex file",
            action="store"
        )

        self.parser.add_argument(
            "--from-url",
            help="Get document or information from url",
            default=None,
            action="store"
        )

        self.parser.add_argument(
            "--from-doi",
            help="Doi to try to get information from",
            default=None,
            action="store"
        )

        self.parser.add_argument(
            "--auto",
            help="Try to parse information from different sources",
            action="store_true"
        )

    def main(self):
        # TODO: Try to recycle some of this code with command add.
        documents = papis.api.get_documents_in_lib(
            self.get_args().lib,
            self.get_args().search
        )
        document = self.pick(documents)
        if not document: return 0
        data = dict()

        if self.args.auto:
            if 'doi' in document.keys() and not self.args.from_doi:
                self.args.from_doi = document['doi']
            if 'title' in document.keys() and not self.args.from_isbnplus:
                self.args.from_isbnplus = document['title']

        if self.args.from_yaml:
            import yaml
            data.update(yaml.load(open(self.args.from_yaml)))

        if self.args.from_isbnplus:
            doc = self.pick(
                [
                    papis.document.Document(data=d)
                    for d in papis.isbn.get_data(
                        query=self.args.from_isbnplus
                    )
                ]
            )
            if doc:
                data.update(doc.to_dict())

        if self.args.from_bibtex:
            bib_data = papis.bibtex.bibtex_to_dict(self.args.from_bibtex)
            if len(bib_data) > 1:
                self.logger.warning(
                    'Your bibtex file contains more than one entry,'
                    ' I will be taking the first entry'
                )
            data.update(bib_data[0])

        if self.args.from_url:
            url_data = papis.downloaders.utils.get(self.args.from_url)
            data.update(url_data["data"])
            document_paths = url_data["documents_paths"]
            if not len(document_paths) == 0:
                document_path = document_paths[0]
                old_doc = self.pick(document["files"])
                if papis.utils.confirm("Really replace document %s?" % old_doc):
                    new_path = os.path.join(
                        document.get_main_folder(), old_doc
                    )
                    self.logger.debug(
                        "Moving %s to %s" %(document_path, new_path)
                    )
                    shutil.move(document_path, new_path)
        if self.args.from_doi:
            self.logger.debug("Try using doi %s" % self.args.from_doi)
            data.update(papis.utils.doi_to_data(self.args.from_doi))

        document.update(data, self.args.force, self.args.interactive)
        document.save()



"""
This command checks for several attributes in every document.

For example if you want to chekc that every document of your library has valid
files related to it you can just do

::

    papis check --keys files

this will check that every info file has the key files and that every file
listed exists.

You can also define your own ones by inputing comma separated keys, e.g. if you
want to check that every document has files, a valid author and title you would
just hut

::

    papis check --keys files,author,title
"""
import papis.api
import papis.config


class Command(papis.commands.Command):
    def init(self):

        self.parser = self.get_subparsers().add_parser(
            "check",
            help="Check document from a given library"
        )

        self.add_search_argument()

        self.parser.add_argument(
            "--keys", "-k",
            help="Comma separated keys to check",
            default=papis.config.get('check-keys'),
            action="store"
        )

    def main(self):
        documents = papis.api.get_documents_in_lib(
            self.get_args().lib,
            self.get_args().search
        )
        all_ok = True
        doc_ok = True
        self.args.keys = self.args.keys.split(',')
        self.logger.debug(self.args.keys)
        for document in documents:
            doc_ok = True
            for key in self.args.keys:
                if key not in document.keys():
                    all_ok &= False
                    doc_ok &= False
                    print(
                        "%s not found in %s" % (
                            key, document.get_main_folder()
                        )
                    )
                elif not document[key] and document[key] is not False:
                    all_ok &= False
                    print(
                        "%s is ill-defined (%s) in %s" % (
                            key, document[key], document.get_main_folder()
                        )
                    )
                elif key == 'files':
                    all_ok &= document.check_files()
        if not all_ok:
            print("Errors were detected, please fix the info files")
        else:
            print("No errors detected")



"""This module describes which functions are intended to be used by users to
create papis scripts.
"""

from subprocess import call
import logging

logger = logging.getLogger("api")
logger.debug("importing")

import os
import re
import papis.cache
import papis.utils
import papis.commands
import papis.config


def get_lib():
    """Get current library, it either retrieves the library from
    the environment PAPIS_LIB variable or from the command line
    args passed by the user.

    :returns: Library name
    :rtype:  str

    >>> get_lib() == papis.config.get_default_settings(key='default-library')
    True
    >>> set_lib('books')
    >>> get_lib()
    'books'
    """
    return papis.config.get_lib()


def set_lib(library):
    """Set current library, it either sets the library in
    the environment PAPIS_LIB variable or in the command line
    args passed by the user.

    :param library: Name of library or path to a given library
    :type  library: str

    """
    try:
        args = papis.commands.get_args()
        args.lib = library
    except AttributeError:
        os.environ["PAPIS_LIB"] = library


def get_arg(arg, default=None):
    try:
        val = getattr(papis.commands.get_args(), arg)
    except AttributeError:
        try:
            val = os.environ["PAPIS_"+arg.upper()]
        except KeyError:
            val = default
    return val


def get_libraries():
    """Get all libraries declared in the configuration. A library is discovered
    if the ``dir`` key defined in the library section.

    :returns: List of library names
    :rtype: list

    """
    libs = []
    config = papis.config.get_configuration()
    for key in config.keys():
        if "dir" in config[key]:
            libs.append(key)
    return libs


def pick(options, pick_config={}):
    """This is a wrapper for the various pickers that are supported.
    Depending on the configuration different selectors or 'pickers'
    are used.

    :param options: List of different objects. The type of the objects within
        the list must be supported by the pickers. This is the reason why this
        function is difficult to generalize for external picker programs.
    :type  options: list

    :param pick_config: Dictionary with additional configuration for the used
        picker. This depends on the picker.
    :type  pick_config: dict

    :returns: Returns elements of ``options``.
    :rtype: Element(s) of ``options``

    >>> papis.config.set('picktool', 'papis.pick')
    >>> pick(['something'])
    'something'

    """
    # Leave this import here
    import papis.config
    logger.debug("Parsing picktool")
    picker = papis.config.get("picktool")
    if picker == "rofi":
        import papis.gui.rofi
        logger.debug("Using rofi picker")
        return papis.gui.rofi.pick(options, **pick_config)
    elif picker == "vim":
        import papis.gui.vim
        logger.debug("Using vim picker")
        return papis.gui.vim.pick(options, **pick_config)
    elif picker == "papis.pick":
        import papis.pick
        logger.debug("Using papis.pick picker")
        return papis.pick.pick(options, **pick_config)
    else:
        raise Exception("I don't know how to use the picker '%s'" % picker)


def open_file(file_path):
    """Open file using the ``opentool`` key value as a program to
    handle file_path.

    :param file_path: File path to be handled.
    :type  file_path: str

    """
    papis.utils.general_open(file_path, "opentool")


def open_dir(dir_path):
    """Open dir using the ``file-browser`` key value as a program to
    open dir_path.

    :param dir_path: Folder path to be handled.
    :type  dir_path: str

    """
    papis.utils.general_open(dir_path, "file-browser")


def edit_file(file_path):
    """Edit file using the ``editor`` key value as a program to
    handle file_path.

    :param file_path: File path to be handled.
    :type  file_path: str

    """
    papis.utils.general_open(file_path, "editor")


def get_documents_in_dir(directory, search=""):
    """Get documents contained in the given folder with possibly a search
    string.

    :param directory: Folder path.
    :type  directory: str

    :param search: Search string
    :type  search: str

    :returns: List of filtered documents.
    :rtype: list

    >>> docs = get_documents_in_dir('non/existent/path')
    >>> len(docs)
    0

    """
    return papis.utils.get_documents(directory, search)


def get_documents_in_lib(library=None, search=""):
    """Get documents contained in the given library with possibly a search
    string.

    :param library: Library name.
    :type  library: str

    :param search: Search string
    :type  search: str

    :returns: List of filtered documents.
    :rtype: list

    """
    directory = library if os.path.exists(library) \
        else papis.config.get("dir", section=library)
    return papis.api.get_documents_in_dir(directory, search)


def clear_lib_cache(lib=None):
    """Clear cache associated with a library. If no library is given
    then the current library is used.

    :param lib: Library name.
    :type  lib: str

    """
    lib = papis.api.get_lib() if lib is None else lib
    directory = papis.config.get("dir", section=lib)
    papis.cache.clear(directory)



import re
import papis.downloaders.base


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(
            self, url, name="scitationaip"
        )
        self.expected_document_format = 'pdf'

    @classmethod
    def match(cls, url):
        # http://aip.scitation.org/doi/10.1063/1.4873138
        if re.match(r".*(aip|aapt).scitation.org.*", url):
            return Downloader(url)
        else:
            return False

    def get_doi(self):
        mdoi = re.match(r'.*/doi/(.*/[^?&%^$]*).*', self.get_url())
        if mdoi:
            doi = mdoi.group(1).replace("abs/", "").replace("full/", "")
            return doi
        else:
            return None

    def get_document_url(self):
        # http://aip.scitation.org/doi/pdf/10.1063/1.4873138
        durl = "http://aip.scitation.org/doi/pdf/%s" % self.get_doi()
        self.logger.debug("[doc url] = %s" % durl)
        return durl

    def get_bibtex_url(self):
        url = "http://aip.scitation.org/action/downloadCitation"\
              "?format=bibtex&cookieSet=1&doi=%s" % self.get_doi()
        self.logger.debug("[bibtex url] = %s" % url)
        return url



import os
import sys
import shutil
import papis.utils
import papis.config
import papis.bibtex


class Document(object):

    """Class implementing the entry abstraction of a document in a library.
    It is basically a python dictionary with more methods.
    """

    subfolder = ""
    _infoFilePath = ""

    def __init__(self, folder=None, data=None):
        self._keys = []
        self._folder = folder
        if folder is not None:
            self._infoFilePath = \
                os.path.join(folder, papis.utils.get_info_file_name())
            self.load()
            self.subfolder = self.get_main_folder()\
                                 .replace(os.environ["HOME"], "")\
                                 .replace("/", " ")
        if data is not None:
            self.update(data)

    def __delitem__(self, key):
        """Deletes property from document, e.g. ``del doc['url']``.
        :param key: Name of the property.
        :type  key: str
        """
        self._keys.pop(self._keys.index(key))
        delattr(self, key)

    def __setitem__(self, key, value):
        """Sets property to value from document, e.g. ``doc['url'] =
        'www.gnu.org'``.
        :param key: Name of the property.
        :type  key: str
        :param value: Value of the parameter
        :type  value: str,int,float,list
        """
        self._keys.append(key)
        setattr(self, key, value)

    def __getitem__(self, key):
        """Gets property to value from document, e.g. ``a = doc['url']``.
        If the property `key` does not exist, then the empy string is returned.

        :param key: Name of the property.
        :type  key: str
        :returns: Value of the property
        :rtype:  str,int,float,list
        """
        return getattr(self, key) if hasattr(self, key) else ""

    def get_main_folder(self):
        """Get full path for the folder where the document and the information
        is stored.
        :returns: Folder path
        """
        return self._folder

    def get_main_folder_name(self):
        """Get main folder name where the document and the information is
        stored.
        :returns: Folder name
        """
        return os.path.basename(self._folder)

    def has(self, key):
        """Check if the information file has some key defined.

        :param key: Key name to be checked
        :returns: True/False
        """
        return key in self.keys()

    def check_files(self):
        """Check for the exsitence of the document's files
        :returns: False if some file does not exist, True otherwise
        :rtype:  bool
        """
        for f in self.get_files():
            # self.logger.debug(f)
            if not os.path.exists(f):
                print("** Error: %s not found in %s" % (
                    f, self.get_main_folder()))
                return False
            else:
                return True

    def rm_file(self, filepath):
        """Remove file from document, it also removes the entry in `files`

        :filepath: Full file path for file
        """
        basename = os.path.basename(filepath)
        if basename not in self['files']:
            raise Exception("File %s not tracked by document" % basename)
        os.remove(filepath)
        self['files'].pop(self['files'].index(basename))

    def rm(self):
        """Removes document's folder, effectively removing it from the library.
        """
        shutil.rmtree(self.get_main_folder())

    def save(self):
        """Saves the current document's information into the info file.
        """
        import yaml
        fd = open(self._infoFilePath, "w+")
        structure = dict()
        for key in self.keys():
            structure[key] = self[key]
        # self.logger.debug("Saving %s " % self.get_info_file())
        yaml.dump(structure, fd, default_flow_style=False)
        fd.close()

    def to_json(self):
        """Export information into a json string
        :returns: Json formatted info file
        :rtype:  str
        """
        import json
        return json.dumps(self.to_dict())

    def to_dict(self):
        """Gets a python dictionary with the information of the document
        :returns: Python dictionary
        :rtype:  dict
        """
        result = dict()
        for key in self.keys():
            result[key] = self[key]
        return result

    @classmethod
    def get_vcf_template(cls):
        return """\
first_name: null
last_name: null
org:
- null
email:
    work: null
    home: null
tel:
    cell: null
    work: null
    home: null
adress:
    work: null
    home: null"""

    def to_vcf(self):
        # TODO: Generalize using the doc variable.
        if not papis.config.in_mode("contact"):
            # self.logger.error("Not in contact mode")
            sys.exit(1)
        text = \
            """\
BEGIN:VCARD
VERSION:4.0
FN:{doc[first_name]} {doc[last_name]}
N:{doc[last_name]};{doc[first_name]};;;""".format(doc=self)
        for contact_type in ["email", "tel"]:
            text += "\n"
            text += "\n".join([
                "{contact_type};TYPE={type}:{tel}"\
                .format(
                    contact_type=contact_type.upper(),
                    type=t.upper(),
                    tel=self[contact_type][t]
                    )
                for t in self[contact_type].keys() \
                if self[contact_type][t] is not None
            ])
        text += "\n"
        text += "END:VCARD"
        return text

    def to_bibtex(self):
        """Create a bibtex string from document's information
        :returns: String containing bibtex formating
        :rtype:  str
        """
        bibtexString = ""
        bibtexType = ""
        # First the type, article ....
        if "type" in self.keys():
            if self["type"] in papis.bibtex.bibtex_types:
                bibtexType = self["type"]
        if not bibtexType:
            bibtexType = "article"
        if not self["ref"]:
            ref = os.path.basename(self.get_main_folder())
        else:
            ref = self["ref"]
        bibtexString += "@%s{%s,\n" % (bibtexType, ref)
        for bibKey in papis.bibtex.bibtex_keys:
            if bibKey in self.keys():
                bibtexString += "\t%s = { %s },\n" % (bibKey, self[bibKey])
        bibtexString += "}\n"
        return bibtexString

    def update(self, data, force=False, interactive=False):
        """Update document's information from an info dictionary.

        :param data: Dictionary with key and values to be updated
        :type  data: dict
        :param force: If True, the update turns into a replace, i.e., it
            replaces the old value by the new value stored in data.
        :type  force: bool
        :param interactive: If True, it will ask for user's input every time
            that the values differ.
        :type  interactive: bool

        """
        for key in data.keys():
            if self[key] != data[key]:
                if force:
                    self[key] = data[key]
                elif interactive:
                    confirmation = \
                        papis.utils.confirm(
                            "(%s conflict) Replace '%s' by '%s'?" % (
                                key, self[key], data[key]
                            )
                        )
                    if confirmation:
                        self[key] = data[key]
                elif self[key] is None or self[key] == '':
                    self[key] = data[key]

    def get_info_file(self):
        """Get full path for the info file
        :returns: Full path for the info file
        :rtype: str
        """
        return self._infoFilePath

    def get_files(self):
        """Get the files linked to the document, if any.

        :returns: List of full file paths
        :rtype:  list
        """
        files = self["files"] if isinstance(self["files"], list) \
            else [self["files"]]
        result = []
        for f in files:
            result.append(os.path.join(self.get_main_folder(), f))
        return result

    def keys(self):
        """Returns the keys defined for the document.

        :returns: Keys for the document
        :rtype:  list
        """
        return self._keys

    def dump(self):
        """Return information string without any obvious format
        :returns: String with document's information
        :rtype:  str

        """
        string = ""
        for i in self.keys():
            string += str(i)+":   "+str(self[i])+"\n"
        return string

    def load(self):
        """Load information from info file
        """
        import yaml
        # TODO: think about if it's better to raise an exception here
        # TODO: if no info file is found
        try:
            fd = open(self._infoFilePath, "r")
        except:
            return False
        structure = yaml.load(fd)
        fd.close()
        for key in structure:
            self[key] = structure[key]



import os
import logging
import urllib.request
import papis.config
import papis.utils


class Downloader(object):

    """This is the base class for every downloader.
    """

    def __init__(self, url="", name=""):
        self.url = url
        self.name = name or os.path.basename(__file__)
        self.logger = logging.getLogger("downloaders:"+self.name)
        self.bibtex_data = None
        self.document_data = None
        self.logger.debug("[url] = %s" % url)
        self.expected_document_format = None

    def __repr__(self):
        return self.get_name()

    @classmethod
    def match(url):
        """This method should be called to know if a given url matches
        the downloader or not.

        For example, a valid match for archive would be:
        .. code:: python

            return re.match(r".*arxiv.org.*", url)

        it will return something that is true if it matches and something
        falsely otherwise.

        :param url: Url where the document should be retrieved from.
        :type  url: str
        """
        raise NotImplementedError(
            "Matching url not implemented for this downloader"
        )

    def get_bibtex_url(self):
        """It returns the urls that is to be access to download
        the bibtex information. It has to be implemented for every
        downloader, or otherwise it will raise an exception.

        :returns: Bibtex url
        :rtype:  str
        """
        raise NotImplementedError(
            "Getting bibtex url not implemented for this downloader"
        )

    def get_bibtex_data(self):
        """Get the bibtex_data data if it has been downloaded already
        and if not download it and return the data in utf-8 format.

        :returns: Bibtex data in utf-8 format
        :rtype:  str
        """
        if not self.bibtex_data:
            self.download_bibtex()
        return self.bibtex_data

    def download_bibtex(self):
        """Bibtex downloader, it should try to download bibtex information from
        the url provided by ``get_bibtex_url``.

        It sets the ``bibtex_data`` attribute if it succeeds.

        :returns: Nothing
        :rtype:  None
        """
        url = self.get_bibtex_url()
        if not url:
            return False
        data = urllib.request.urlopen(url)\
            .read()\
            .decode('utf-8')
        self.bibtex_data = data

    def get_document_url(self):
        """It returns the urls that is to be access to download
        the document information. It has to be implemented for every
        downloader, or otherwise it will raise an exception.

        :returns: Document url
        :rtype:  str
        """
        raise NotImplementedError(
            "Getting bibtex url not implemented for this downloader"
        )

    def get_doi(self):
        """It returns the doi of the document, if it is retrievable.
        It has to be implemented for every downloader, or otherwise it will
        raise an exception.

        :returns: Document doi
        :rtype:  str
        """
        raise NotImplementedError(
            "Getting document url not implemented for this downloader"
        )

    def get_document_data(self):
        """Get the document_data data if it has been downloaded already
        and if not download it and return the data in binary format.

        :returns: Document data in binary format
        :rtype:  str
        """
        if not self.document_data:
            self.download_document()
        return self.document_data

    def download_document(self):
        """Document downloader, it should try to download document information from
        the url provided by ``get_document_url``.

        It sets the ``document_data`` attribute if it succeeds.

        :returns: Nothing
        :rtype:  None
        """
        self.logger.debug("Downloading document")
        url = self.get_document_url()
        if not url:
            return False
        request = urllib.request.Request(
            url,
            headers={
                'User-Agent': papis.config.get('user-agent')
            }
        )
        data = urllib.request.urlopen(request).read()
        self.document_data = data

    def set_url(self, url):
        """Url setter for Downloader

        :param url: String containing a valid url
        :type  url: str
        :returns: Downloader object
        """
        self.url = url
        return self

    def get_url(self):
        """Url getter for Downloader
        :returns: Main url of the Downloader
        :rtype:  str
        """
        return self.url

    def check_document_format(self):
        """Check if the downloaded document has the filetype that the
        downloader expects. If the downloader does not expect any special
        filetype, accept anything because there is no way to know if it is
        correct.

        :returns: True if it is of the right type, else otherwise
        :rtype:  bool
        """
        if self.expected_document_format is None:
            return True
        result = papis.utils.file_is(
            self.get_document_data(),
            self.expected_document_format
        )
        if not result:
            self.logger.warning(
                "The downloaded data does not seem to be of"
                "the correct type (%s)" % self.expected_document_format
            )
        return result

    def get_name(self):
        """Get name of the downloader
        :returns: Name

        """
        return self.name



"""
This command is to list contents of a library.

Examples
^^^^^^^^

 - List all document files associated will all entries:

    .. code:: bash

        papis list --file

    .. raw:: HTML

        <script type="text/javascript"
        src="https://asciinema.org/a/XwD0ZaUORoOonwDw4rXoQDkjZ.js"
        id="asciicast-XwD0ZaUORoOonwDw4rXoQDkjZ" async></script>

 - List all document year and title with custom formatting:

    .. code:: bash

        papis list --format '{doc[year]} {doc[title]}'

    .. raw:: HTML

        <script type="text/javascript"
        src="https://asciinema.org/a/NZ8Ii1wWYPo477CIL4vZhUqOy.js"
        id="asciicast-NZ8Ii1wWYPo477CIL4vZhUqOy" async></script>

 - List all documents according to the bibitem formatting (stored in a template
   file ``bibitem.template``):

    .. code:: bash

        papis list --template bibitem.template

    .. raw:: HTML

        <script type="text/javascript"
        src="https://asciinema.org/a/QZTBZ3tFfyk9WQuJ9WWB2UpSw.js"
        id="asciicast-QZTBZ3tFfyk9WQuJ9WWB2UpSw" async></script>

"""

import papis
import os
import sys
import papis.utils
import papis.downloaders.utils


class Command(papis.commands.Command):
    def init(self):

        self.parser = self.get_subparsers().add_parser(
            "list",
            help="List documents from a given library"
        )

        self.add_search_argument()

        self.parser.add_argument(
            "-i",
            "--info",
            help="Show the info file name associated with the document",
            default=False,
            action="store_true"
        )

        self.parser.add_argument(
            "-f",
            "--file",
            help="Show the file name associated with the document",
            action="store_true"
        )

        self.parser.add_argument(
            "-d",
            "--dir",
            help="Show the folder name associated with the document",
            action="store_true"
        )

        self.parser.add_argument(
            "--format",
            help="List entries using a custom papis format, e.g."
                " '{doc[year] {doc[title]}",
            default=None,
            action="store"
        )

        self.parser.add_argument(
            "--template",
            help="Template file containing a papis format to list entries",
            default=None,
            action="store"
        )

        self.parser.add_argument(
            "-p",
            "--pick",
            help="Pick the document instead of listing everything",
            action="store_true"
        )

        self.parser.add_argument(
            "--downloaders",
            help="List available downloaders",
            action="store_true"
        )

    def main(self):
        if self.args.template:
            if not os.path.exists(self.args.template):
                self.logger.error(
                    "Template file %s not found" % self.args.template
                )
                return 1
            fd = open(self.args.template)
            self.args.format = fd.read()
            fd.close()
        if self.args.downloaders:
            for downloader in \
               papis.downloaders.utils.getAvailableDownloaders():
                print(downloader)
            return 0

        documents = papis.api.get_documents_in_lib(
            self.get_args().lib,
            self.get_args().search
        )

        if self.args.pick:
            documents = [self.pick(documents)]
        for document in documents:
            if self.args.file:
                for f in document.get_files():
                    print(f)
            elif self.args.info:
                print(
                    os.path.join(
                        document.get_main_folder(),
                        document.get_info_file()
                    )
                )
            elif self.args.format:
                print(
                    papis.utils.format_doc(self.args.format, document)
                )
            else:
                print(document.get_main_folder())



import papis
import os
import sys
import papis.api
import papis.utils
import papis.pick
import papis.config


class Command(papis.commands.Command):
    def init(self):

        self.parser = self.get_subparsers().add_parser(
            "edit",
            help="Edit document information from a given library"
        )

        self.add_search_argument()

        self.parser.add_argument(
            "-n",
            "--notes",
            help="Open notes document",
            action="store_true"
        )

    def main(self):

        documents = papis.api.get_documents_in_lib(
            self.get_args().lib,
            self.args.search
        )
        document = self.pick(documents)
        if not document: return 0

        if self.args.notes:
            self.logger.debug("Editing notes")
            if not document.has("notes"):
                self.logger.warning(
                    "The document selected has no notes attached,"\
                    " creating one..."
                )
                document["notes"] = papis.config.get("notes-name")
                document.save()
            notesPath = os.path.join(
                document.get_main_folder(),
                document["notes"]
            )
            if not os.path.exists(notesPath):
                self.logger.debug("Creating %s" % notesPath)
                open(notesPath, "w+").close()
            papis.api.edit_file(notesPath)
        else:
            papis.api.edit_file(document.get_info_file())



import re
import papis.downloaders.base


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(
            self, url, name="iopscience"
        )
        self.expected_document_format = 'pdf'

    @classmethod
    def match(cls, url):
        """
        >>> Downloader.match(\
                'http://iopscience.iop.org/article/10.1088/0305-4470/24/2/004'\
            ) is False
        False
        >>> Downloader.match(\
                'blah://iop.org/!@#!@$!%!@%!$che.6b00559'\
            ) is False
        True
        >>> Downloader.match(\
                'iopscience.iop.com/!@#!@$!%!@%!$chemed.6b00559'\
            ) is False
        True
        """
        if re.match(r".*iopscience.iop.org.*", url):
            return Downloader(url)
        else:
            return False

    def get_doi(self):
        mdoi = re.match(r'.*annualreviews.org/doi/(.*)', self.get_url())
        if mdoi:
            doi = mdoi.group(1).replace("abs/", "").replace("full/", "")
            self.logger.debug("[doi] = %s" % doi)
            return doi
        else:
            self.logger.error("Doi not found!!")

    def get_document_url(self):
        # http://iopscience.iop.org/article/10.1088/0305-4470/24/2/004/pdf
        durl = self.get_url()+"/pdf"
        self.logger.debug("[doc url] = %s" % durl)
        return durl

    def getAritcleId(self):
        """Get article's id for IOP
        :returns: Article id
        """
        url = self.get_url()
        m = re.match(r"^.*iop.org/[^/]+/[^/]+/(.*)$", url)
        if not m:
            self.logger.error("Could not retrieve articleId from url")
            return None
        articleId = m.group(1)
        self.logger.debug("[doc articleId] = %s" % articleId)
        return articleId

    def get_bibtex_url(self):
        # http://iopscience.iop.org/export?articleId=0305-4470/24/2/004&exportFormat=iopexport_bib&exportType=abs&navsubmit=Export%2Babstract
        articleId = self.getAritcleId()
        url = "http://iopscience.iop.org/export?articleId=%s&exportFormat=iopexport_bib&exportType=abs&navsubmit=Export%%2Babstract" % articleId
        self.logger.debug("[bibtex url] = %s" % url)
        return url

# vim-run: python3 %



import logging

import tempfile
import papis.bibtex
import papis.config

logger = logging.getLogger("downloader")


def getAvailableDownloaders():
    import papis.downloaders.aps
    import papis.downloaders.acs
    import papis.downloaders.arxiv
    import papis.downloaders.scitationaip
    import papis.downloaders.annualreviews
    import papis.downloaders.iopscience
    import papis.downloaders.libgen
    import papis.downloaders.get
    return [
        papis.downloaders.aps.Downloader,
        papis.downloaders.acs.Downloader,
        papis.downloaders.arxiv.Downloader,
        papis.downloaders.scitationaip.Downloader,
        papis.downloaders.annualreviews.Downloader,
        papis.downloaders.iopscience.Downloader,
        papis.downloaders.libgen.Downloader,
        papis.downloaders.get.Downloader,
    ]


def get_downloader(url, downloader=False):
    """Get downloader object. If only a url is given, the url is matched
    against the match method of the downloaders.

    :param url: Url of the document
    :type  url: str
    :param downloader: Name of the downloader, if any.
    :type  downloader: str

    """
    if not downloader:
        for downloader in getAvailableDownloaders():
            result = downloader.match(url)
            if result:
                return result
    else:
        try:
            mod = __import__(
                'papis.downloaders.%s' % downloader,
                fromlist=['Downloader']
            )
            return getattr(mod, 'Downloader')(url)
        except ImportError:
            logger.error("No downloader named %s" % downloader)
    return False


def download_document_from_doi(doi):
    """Download a document knowing the doi number

    :param doi: Doi number
    :type  doi: string
    :returns: Path to downloaded local document

    """
    tempf = tempfile.mktemp()
    external_downloader = papis.config.get(
        "doc-doi-downloader"
    )



def get(url, data_format="bibtex", expected_doc_format=None):
    logger.debug("Attempting to retrieve from url")
    result = {
        "data": dict(),
        "doi": None,
        "documents_paths": []
    }
    downloader = get_downloader(url)
    if not downloader:
        logger.warning(
            "No matching Downloader for the url %s found" % url
        )
        return result
    logger.debug("Using downloader %s" % downloader)
    if downloader.expected_document_format is None and \
            expected_doc_format is not None:
        downloader.expected_document_format = expected_doc_format
    try:
        doi = downloader.get_doi()
    except:
        logger.debug("Doi not found from url...")
    if data_format == "bibtex":
        try:
            bibtex_data = downloader.get_bibtex_data()
            if bibtex_data:
                result["data"] = papis.bibtex.bibtex_to_dict(
                    bibtex_data
                )
                result["data"] = result["data"][0] \
                    if len(result["data"]) > 0 else dict()
        except NotImplementedError:
            pass
    try:
        doc_data = downloader.get_document_data()
    except NotImplementedError:
        doc_data = False
    if doc_data:
        if downloader.check_document_format():
            result["documents_paths"].append(tempfile.mktemp())
            logger.debug(
                "Saving downloaded data in %s" % result["documents_paths"][-1]
            )
            with open(result["documents_paths"][-1], "wb+") as fd:
                fd.write(doc_data)
    return result



"""
This command is useful if your library is itself a git repository.
You can use this command to issue git commands in your library
repository without having to change your current directory.

Here are some examples of its usage:

    - Check the status of the library repository:

    .. code::

        papis git status

    - Commit all changes:

    .. code::

        papis git commit -a
"""
import papis.commands
import argparse


class Command(papis.commands.Command):

    def init(self):

        self.parser = self.get_subparsers().add_parser(
            "git",
            help="Run a git command in the library folder"
        )

        self.parser.add_argument(
            "commands",
            help="Commands",
            default="",
            nargs=argparse.REMAINDER,
            action="store"
        )

    def main(self):
        run = papis.commands.get_commands("run")
        run.set_args(self.get_args())
        run.set_commands(["git"] + self.args.commands)
        run.main()



"""
The export command is useful to work with other programs such as bibtex.

Some examples of its usage are:

    - Export one of the documents matching the author with einstein to bibtex:

    .. code::

        papis export --bibtex 'author = einstein'

    or export all of them

    .. code::

        papis export --bibtex --all 'author = einstein'

    - Export all documents to bibtex and save them into a ``lib.bib`` file

    .. code::

        papis export --all --bibtex --out lib.bib

    - Export a folder of one of the documents matching the word ``krebs``
      into a folder named, ``interesting-document``

    .. code::

        papis export --folder --out interesting-document krebs

    this will create the folder ``interesting-document`` containing the
    ``info.yaml`` file, the linked documents and a ``bibtex`` file for
    sharing with other people.

"""
import papis
import os
import sys
import shutil
import papis.utils


class Command(papis.commands.Command):
    def init(self):

        self.parser = self.get_subparsers().add_parser(
            "export",
            help="""Export a document from a given library"""
        )

        self.add_search_argument()

        self.parser.add_argument(
            "--yaml",
            help="Export into yaml",
            default=False,
            action="store_true"
        )

        self.parser.add_argument(
            "--bibtex",
            help="Export into bibtex",
            default=False,
            action="store_true"
        )

        self.parser.add_argument(
            "--json",
            help="Export into json",
            default=False,
            action="store_true"
        )

        self.parser.add_argument(
            "--folder",
            help="Export document folder to share",
            default=False,
            action="store_true"
        )

        self.parser.add_argument(
            "--no-bibtex",
            help="When exporting to a folder, do not include the bibtex",
            default=False,
            action="store_true"
        )

        self.parser.add_argument(
            "-o",
            "--out",
            help="Outfile or outdir",
            default="",
            action="store"
        )

        self.parser.add_argument(
            "-t",
            "--text",
            help="Text formated reference",
            action="store_true"
        )

        self.parser.add_argument(
            "-a", "--all",
            help="Export all without picking",
            action="store_true"
        )

        self.parser.add_argument(
            "--vcf",
            help="Export contact to vcf format",
            action="store_true"
        )

    def main(self):

        documents = papis.api.get_documents_in_lib(
            self.get_args().lib,
            self.get_args().search
        )

        if not self.args.all:
            document = self.pick(documents)
            if not document: return 0
            documents = [document]

        if self.args.out and not self.get_args().folder:
            self.args.out = open(self.get_args().out, 'a+')

        if not self.args.out and not self.get_args().folder:
            self.args.out = sys.stdout

        if self.args.json:
            import json
            return self.args.out.write(
                json.dumps([document.to_dict() for document in documents])
            )

        if self.args.yaml:
            import yaml
            return self.args.out.write(
                yaml.dump_all([document.to_dict() for document in documents])
            )

        for document in documents:
            if self.args.bibtex:
                self.args.out.write(document.to_bibtex())
            if self.args.text:
                text_format = papis.config.get('export-text-format')
                text = papis.utils.format_doc(text_format, document)
                self.args.out.write(text)
            elif self.args.folder:
                folder = document.get_main_folder()
                outdir = self.args.out or document.get_main_folder_name()
                if not len(documents) == 1:
                    outdir = os.path.join(
                        outdir, document.get_main_folder_name())
                shutil.copytree(folder, outdir)
                if not self.args.no_bibtex:
                    open(
                        os.path.join(outdir, "info.bib"),
                        "a+"
                    ).write(document.to_bibtex())
            elif self.args.vcf:
                self.args.out.write(document.to_vcf())
            else:
                pass




try:
    # py3
    import tkinter as tk
except:
    # py2
    import Tkinter as tk

from papis.gui.tk.main import PapisWidget
from papis.gui.tk.main import Gui
from papis.gui.tk.prompt import Prompt
from papis.gui.tk.list import PapisList



"""This module implements custom exceptions used to make the code more
readable.
"""


class SettingNotRegistered(Exception):
    """This exception is when a setting is to be retrieved that has not been
    considered beforehand in the code and registered accordingly, thus giving a
    default value to it"""
    pass


class DefaultSettingValueMissing(Exception):
    """This exception is when a setting's value has no default value.
    """

    def __init__(self, key):
        message = "\nValue for '%s' is not at all registered and known" % (key)
        super(DefaultSettingValueMissing, self).__init__(message)



from papis.gui.tk import PapisWidget
from papis.gui.tk import tk
import papis.api
import papis.config
import papis.utils


class ListEntry(tk.Frame, PapisWidget):

    def __init__(self, master, text):
        tk.Frame.__init__(self, master)
        PapisWidget.__init__(self)
        self["justify"] = tk.LEFT
        self["padx"] = 10
        self["font"] = "12"
        self["borderwidth"] = 1
        self["fg"] = self.get_config("entry-fg")
        self["anchor"] = tk.W
        self["activeforeground"] = self.get_config("activeforeground")
        self["activebackground"] = self.get_config("activebackground")
        self.text = text
        self.lines = text.split("\n")
        self["height"] = self.lines
        self.lines_label = []
        for line in self.lines:
            self.lines_label.append(
                tk.Label(
                    self
                )
            )


class PapisList(tk.Frame, PapisWidget):

    index_draw_first = 0
    index_draw_last = 0
    key = None
    selected = None
    index = 0
    entries_drawning = False
    doc_primitive_height = 10
    matched_indices = []
    documents_lbls = []

    def __init__(self, master, *args, **kwargs):
        tk.Frame.__init__(self, master, *args, **kwargs)
        PapisWidget.__init__(self)
        self["bg"] = self.master["bg"]

    def set_bindings(self):
        self.logger.debug("Setting bindings")
        self.noremap("<Control-l>", self.redraw_screen)
        self.master.prompt.cmap("<KeyPress>", self.filter_and_draw)
        self.master.prompt.cmap("<Control-n>", self.move_down)
        self.master.prompt.cmap("<Control-p>", self.move_up)
        self.master.prompt.cmap("<Control-o>", self.open)
        self.master.prompt.cmap("<Control-i>", self.print_info)
        self.bindings = [
            (self.get_config("move_down"), "move_down"),
            (self.get_config("move_up"), "move_up"),
            (self.get_config("open"), "open"),
            (self.get_config("edit"), "edit"),
            (self.get_config("move_top"), "move_top"),
            (self.get_config("move_bottom"), "move_bottom"),
            (self.get_config("print_info"), "print_info"),
            (self.get_config("half_down"), "half_down"),
            (self.get_config("half_up"), "half_up"),
            (self.get_config("scroll_down"), "scroll_down"),
            (self.get_config("scroll_up"), "scroll_up"),
            ("<Down>", "move_down"),
            ("<Up>", "move_up"),
        ]
        for bind in self.bindings:
            key = bind[0]
            name = bind[1]
            self.master.nmap(key, getattr(self, name))

    def get_matched_indices(self, force=False):
        if not self.master.prompt.changed() and not force:
            return self.matched_indices
        self.logger.debug("Indexing")
        command = self.master.prompt.get_command()
        match_format = self.get_config("match-format")
        indices = list()
        for i, doc in enumerate(self.documents):
            if papis.utils.match_document(doc, command, match_format):
                indices.append(i)
        self.matched_indices = indices
        return indices

    def filter_and_draw(self, event=None):
        indices = self.get_matched_indices()
        self.undraw_documents_labels()
        self.draw_documents_labels(indices)

    def set_documents(self, docs):
        self.documents = docs

    def scroll(self, direction):
        self.undraw_documents_labels()
        if direction == "down":
            self.index_draw_first += 1
        else:
            if self.index_draw_first > 0:
                self.index_draw_first -= 1
        self.update_selection_index()
        self.draw_documents_labels()

    def scroll_down(self, event=None):
        self.scroll("down")

    def scroll_up(self, event=None):
        self.scroll("up")

    def half_up(self, event=None):
        self.logger.debug("Half up")
        print("TODO")

    def half_down(self, event=None):
        self.logger.debug("Half down")
        print("TODO")

    def move_top(self, event=None):
        self.logger.debug("Moving to top")
        self.index_draw_first = 0
        self.index = self.index_draw_first
        self.redraw_documents_labels()

    def move_bottom(self, event=None):
        self.logger.debug("Moving to bottom")
        self.index_draw_first = len(self.get_matched_indices())-1
        self.index = self.index_draw_first
        self.redraw_documents_labels()

    def move_down(self, event=None):
        self.move("down")

    def move_up(self, event=None):
        self.move("up")

    def draw_selection(self, event=None):
        indices = self.get_matched_indices()
        if not len(indices):
            return False
        if self.get_selected() is not None:
            self.get_selected().configure(state="normal")
        self.update_selection_index()
        self.set_selected(self.documents_lbls[indices[self.index]])
        self.get_selected().configure(state="active")

    def get_selected(self):
        return self.selected

    def get_selected_doc(self):
        return self.selected.doc

    def set_selected(self, doc_lbl):
        self.selected = doc_lbl

    def set_documents_labels(self):
        self.update_drawing_indices()
        font_size = self.get_config("entry-font-size")
        font_name = self.get_config("entry-font-name")
        font_style = self.get_config("entry-font-style")
        font_lines = self.get_config("entry-lines")
        font = (font_name, font_size, font_style)
        doc_primitive_height = int(font_size)*int(font_lines)
        # number_of_entries = self.winfo_height()/doc_primitive_height
        pady = int(abs(self.winfo_height() - int(font_size)*int(font_lines))/2)
        for doc in self.documents:
            self.documents_lbls.append(
                tk.Label(
                    text=papis.utils.format_doc(
                        self.get_config("header-format"),
                        doc
                    ),
                    justify=tk.LEFT,
                    padx=10,
                    pady=pady,
                    height=font_lines,
                    font=font,
                    borderwidth=1,
                    fg=self.get_config("entry-fg"),
                    anchor=tk.W,
                    activeforeground=self.get_config("activeforeground"),
                    activebackground=self.get_config("activebackground")
                )
            )
            setattr(self.documents_lbls[-1], "doc", doc)

    def redraw_documents_labels(self):
        self.undraw_documents_labels()
        self.draw_documents_labels()

    def undraw_documents_labels(self):
        if self.entries_drawning:
            return False
        if not len(self.documents_lbls):
            return False
        for doc in self.documents_lbls:
            doc.pack_forget()

    def draw_documents_labels(self, indices=[]):
        if self.entries_drawning:
            return False
        else:
            self.logger.debug("Drawing")
            self.entries_drawning = True
        if not len(indices):
            indices = self.get_matched_indices()
        colors = (
            self.get_config("entry-bg-odd"),
            self.get_config("entry-bg-pair"),
        )
        self.update_height()
        self.update_drawing_indices()
        for i in range(self.index_draw_first, self.index_draw_last):
            if i >= len(indices):
                break
            doc = self.documents_lbls[indices[i]]
            doc["bg"] = colors[i % 2]
            doc.pack(
                fill=tk.X
            )
        self.logger.debug("Drawing done")
        self.entries_drawning = False
        self.draw_selection()

    def redraw_screen(self, event=None):
        self.redraw_documents_labels()

    def update_drawing_indices(self):
        label_number = self.get_config("labels-per-page")
        self.doc_primitive_height = int(self["height"]/label_number)
        self.index_draw_last = self.index_draw_first +\
            int(self["height"]/self.doc_primitive_height)
        self.logger.debug("label_h %s" % self.doc_primitive_height)
        self.logger.debug("i_draw_last %s" % self.index_draw_last)

    def update_selection_index(self):
        indices = self.get_matched_indices()
        if self.index < self.index_draw_first:
            self.index = self.index_draw_first
        if self.index > self.index_draw_last:
            self.index = self.index_draw_last-1
        if self.index > len(indices)-1:
            self.index = len(indices)-1

    def move(self, direction):
        indices = self.get_matched_indices()
        if direction == "down":
            if self.index < len(indices)-1:
                self.index += 1
        if direction == "up":
            if self.index > 0:
                self.index -= 1
        if self.index > self.index_draw_last-1:
            self.scroll_down()
        if self.index < self.index_draw_first:
            self.scroll_up()
        self.logger.debug(
            "index = %s in (%s , %s)"
            % (self.index, self.index_draw_first, self.index_draw_last)
        )
        self.draw_selection()

    def open(self, event=None):
        doc = self.get_selected_doc()
        papis.api.open_file(
            doc.get_files()
        )

    def print_info(self, event=None):
        doc = self.get_selected_doc()
        self.master.prompt.echomsg(
            doc.dump()
        )

    def edit(self, event=None):
        doc = self.get_selected_doc()
        papis.utils.general_open(
            doc.get_info_file(),
            "xeditor",
            default_opener="xterm -e vim",
            wait=True
        )
        doc.load()

    def update_height(self):
        self["height"] = self.master.winfo_height() - \
            self.master.prompt.winfo_height()
        self.logger.debug("Updating height (%s)" % self["height"])
        assert(not self["height"] == 0)

    def init(self):
        self.set_bindings()
        self.logger.debug("Creating labels")
        # force indexing
        self.logger.debug("Forcing indexing...")
        self.get_matched_indices(True)
        self.update_height()
        self.set_documents_labels()
        self.draw_documents_labels()



import re
import urllib.request
import papis.downloaders.base


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(self, url, name="aps")
        self.expected_document_format = 'pdf'

    @classmethod
    def match(cls, url):
        """
        >>> Downloader.match(\
                'blah://pubs.aps.org/doi/abs/10.1021/acs.jchemed.6b00559'\
            ) is False
        False
        >>> Downloader.match(\
                'blah://pubs.aps.org/!@#!@$!%!@%!$che.6b00559'\
            ) is False
        False
        >>> Downloader.match(\
                'aps.com/!@#!@$!%!@%!$chemed.6b00559'\
            ) is False
        True
        """
        if re.match(r".*aps.org.*", url):
            return Downloader(url)
        else:
            return False

    def get_document_url(self):
        # http://whatever.aps.org/whatever/whatever/10.1103/PhysRevLett.115.066402
        # https://journals.aps.org/prl/pdf/10.1103/PhysRevLett.119.030403
        url = self.get_url()
        burl = re.sub(r'(aps.org/[a-z]+)/([a-z]+)', r'\1/pdf', url)
        self.logger.debug("[document url] = %s" % burl)
        return burl

    def get_doi(self):
        # http://whatever.aps.org/whatever/whatever/(10.1103/PhysRevLett.115.066402)
        url = self.get_url()
        burl = re.sub(r'.*(aps.org/[a-z]+/[a-z]+/)', r'', url)
        self.logger.debug("[document doi] = %s" % burl)
        return burl

    def get_bibtex_url(self):
        # http://journals.aps.org/prl/export/10.1103/
        # PhysRevLett.115.066402?type=bibtex&download=true
        # http://journals.aps.org/prl/abstract/10.1103/PhysRevLett.115.066402
        url = self.get_url()
        burl = re.sub(r'(aps.org/[a-z]+)/abstract', r'\1/export', url)\
            + "?type=bibtex&download=true"
        self.logger.debug("[bibtex url] = %s" % burl)
        return burl

    def download_bibtex(self):
        data = urllib.request.urlopen(self.get_bibtex_url())\
            .read()\
            .decode('utf-8')
        self.bibtex_data = data

# vim-run: python3 -m doctest %



# -*-coding:utf-8-*-
# adapted from
# https://github.com/wong2/pick

import curses
import re
import string


__all__ = ['Picker', 'pick']

CTRL_P = 16
CTRL_N = 14
CTRL_H = 8
CTRL_C = 3
CTRL_D = 4
CTRL_U = 21
CTRL_W = 23
KEYS_ENTER = [curses.KEY_ENTER, ord('\n'), ord('\r')]
KEYS_UP = [curses.KEY_UP, CTRL_P]
KEYS_DOWN = [curses.KEY_DOWN, CTRL_N]
KEYS_ERASE = [curses.KEY_BACKSPACE, CTRL_H]
KEYS_QUIT = (CTRL_C, curses.KEY_EXIT, 27)
KEYS_HALF_DOWN = [CTRL_D]
KEYS_HALF_UP = [CTRL_U]
KEYS_DEL_WORD = [CTRL_W]


class Picker(object):
    """The :class:`Picker <Picker>` object

    :param options: a list of options to choose from
    :param title: (optional) a title above options list
    :param indicator: (optional) custom the selection indicator
    :param default_index: (optional) set this if the default
        selected option is not the first one
    """

    def __init__(
            self,
            options,
            title=None,
            indicator='*',
            default_index=0,
            header_filter=lambda x: x,
            body_filter=None,
            match_filter=lambda x: x
            ):

        self.options = options
        self.title = title
        self.indicator = indicator
        self.search = ""
        self.header_filter = header_filter
        self.body_filter = body_filter
        self.match_filter = match_filter

        if default_index >= len(options):
            raise ValueError(
                'default_index should be less than the length of options'
            )

        self.index = default_index
        self.custom_handlers = {}

    def move_up(self):
        self.index -= 1
        if self.index < 0:
            self.index = len(self.options) - 1

    def move_down(self):
        self.index += 1
        if self.index >= len(self.options):
            self.index = 0

    def get_selected(self):
        """return the current selected option
        """
        return self.get_filtered_options()[self.index]

    def get_title_lines(self):
        """Return the main row of the picker, which is normally the search and
        the prompt text.
        """
        return [self.title+self.search]

    def get_option_lines(self):
        headers = []
        index_found = False
        last_index = -1
        for index, option in enumerate(self.get_filtered_options()):
            last_index += 1
            if index == self.index:
                index_found = True
                prefix = self.indicator
            else:
                prefix = len(self.indicator) * ' '
            line = '{0} {1}'.format(prefix, self.header_filter(option))
            headers.append(line)
        if not index_found:
            self.index = last_index
            if len(headers):
                headers[-1] = "{0} {1}"\
                        .format(self.indicator, headers[-1].strip(" "))

        return headers

    def get_search_regex(self):
        """TODO: Docstring for get_search_regex.
        :returns: TODO

        """
        cleaned_search = self.search
        # Clean up ( and )
        cleaned_search = cleaned_search.replace('(', '\\(')\
                                       .replace(')', '\\)')\
                                       .replace('+', '\\+')\
                                       .replace('[', '\\[')\
                                       .replace(']', '\\]')
        return r".*"+re.sub(r"\s+", ".*", cleaned_search)

    def get_filtered_options(self):
        """TODO: Docstring for get_filtered_options.
        :returns: TODO

        """
        new_options = []
        regex = self.get_search_regex()
        for option in self.options:
            if re.match(regex, self.match_filter(option), re.I):
                new_options += [option]
        return new_options

    def get_lines(self):
        """TODO: Docstring for get_filtered_lines.
        :returns: TODO

        """
        title_lines = self.get_title_lines()
        option_lines = self.get_option_lines()
        lines = title_lines + option_lines
        current_line = self.index + len(title_lines) + 1
        return lines, current_line

    def draw(self):
        """draw the curses ui on the screen, handle scroll if needed"""
        self.screen.clear()

        x, y = 1, 0  # start point
        max_y, max_x = self.screen.getmaxyx()
        max_rows = max_y - y  # the max rows we can draw

        lines, current_line = self.get_lines()

        # calculate how many lines we should scroll, relative to the top
        scroll_top = getattr(self, 'scroll_top', 0)
        if current_line <= scroll_top:
            scroll_top = 0
        elif current_line - scroll_top > max_rows:
            scroll_top = current_line - max_rows
        self.scroll_top = scroll_top

        lines_to_draw = lines[scroll_top:scroll_top+max_rows]

        for line in lines_to_draw:
            self.screen.addnstr(y, x, line, max_x-2)
            y += 1

        self.screen.refresh()

    def editSearch(self, c):
        """TODO: Docstring for editSearch.

        :c: TODO
        :returns: TODO

        """
        if c in KEYS_ERASE:
            self.search = self.search[0:-1]
        elif c in KEYS_DEL_WORD:
            self.search = re.sub(r"\s*$", "", self.search)
            self.search = re.sub(r"\s*\b[^ ]+$", "", self.search)
            self.draw()
        else:
            self.search += chr(c)

    def run_loop(self):
        while True:
            self.draw()
            c = self.screen.getch()
            if c in KEYS_UP:
                self.move_up()
            elif c in KEYS_DOWN:
                self.move_down()
            elif c in KEYS_ENTER:
                return self.get_selected()
            elif c in KEYS_QUIT:
                curses.endwin()
                return ""
            elif c in KEYS_HALF_UP:
                max_y, max_x = self.screen.getmaxyx()
                new_index = self.index - int(max_y/4)
                if new_index < 0:
                    new_index = 0
                self.index = new_index
            elif c in KEYS_HALF_DOWN:
                max_y, max_x = self.screen.getmaxyx()
                new_index = self.index + int(max_y/4)
                if new_index >= max_y:
                    new_index = max_y
                self.index = new_index
            elif chr(c) in string.printable or c in KEYS_ERASE + KEYS_DEL_WORD:
                self.editSearch(c)

    def config_curses(self):
        # use the default colors of the terminal
        curses.use_default_colors()
        # hide the cursor
        curses.curs_set(0)

    def _start(self, screen):
        self.screen = screen
        self.config_curses()
        return self.run_loop()

    def start(self):
        if len(self.options) == 0:
            return ""
        if len(self.options) == 1:
            return self.options[0]
        return curses.wrapper(self._start)


def pick(
        options,
        title="Pick: ",
        indicator='>',
        default_index=0,
        header_filter=lambda x: x,
        body_filter=None,
        match_filter=lambda x: x
        ):
    """Construct and start a :class:`Picker <Picker>`.
    """
    picker = Picker(
                options,
                title,
                indicator,
                default_index,
                header_filter,
                body_filter,
                match_filter
                )
    return picker.start()




def get_default_settings():
    return {
        "vim-gui": {
            "help-key"          : "h",
            "open-key"          : "o",
            "edit-key"          : "e",
            "search-key"        : "/",
            "delete-key"        : "dd",
            "open-dir-key"      : "<S-o>",
            "next-search-key"   : "n",
            "prev-search-key"   : "N",
            "header-format" : \
                "Title : {doc[title]}\n"\
                "Author: {doc[author]}\n"\
                "Year  : {doc[year]}\n"\
                "-------\n",
        },
        "tk-gui": {
            "open"         : "o",
            "edit"         : "e",
            "exit"         : "<Control-q>",
            "clear"        : "q",
            "help"         : "h",
            "focus_prompt" : ":",
            "move_down"    : "j",
            "move_up"      : "k",
            "move_top"     : "g",
            "move_bottom"  : "<Shift-G>",
            "print_info"   : "i",
            "half_down"    : "<Control-d>",
            "half_up"      : "<Control-u>",
            "scroll_down"  : "<Control-e>",
            "scroll_up"    : "<Control-y>",
            "prompt-fg"    : "lightgreen",
            "prompt-bg"        : "black",
            # Color of the foreground of an entry
            "entry-fg"         : "grey77",
            # Color of the foreground of an active entry
            "activeforeground" : "gray99",
            # Color of the background of an active entry
            "activebackground" : "#394249",
            "insertbackground" : "red",
            "prompt-font-size" : "14",
            "entry-bg-size"    : "14",
            "entry-font-size"  : "14",
            "entry-font-name"  : "Times",
            "entry-font-style" : "normal",
            "entry-lines"      : "3",
            "entry-bg-odd"     : "#273238",
            "entry-bg-pair"    : "#273238",
            "cursor"           : "xterm",
            "height"           : 1,
            "labels-per-page"  : 6,
            "borderwidth"      : -1,
            "window-width"     : "1200",
            "window-bg"        : "#273238",
            "window-height"    : "700",
            "match-format" : \
                "{doc[tags]}{doc.subfolder}{doc[title]}{doc[author]}{doc[year]}",
            "header-format"    : \
                "{doc[title]}\n"\
                "{doc[empty]}   {doc[author]}\n"\
                "({doc[year]:->4})",
        },
        "rofi-gui": {
            "key-quit"          : "Alt+q",
            "key-edit"          : "Alt+e",
            "key-delete"        : "Alt+d",
            "key-help"          : "Alt+h",
            "key-open-stay"     : "Alt+o",
            "key-normal-window" : "Alt+w",
            "key-browse"        : "Alt+u",
            "key-open"          : "Enter",
            "eh"                : 3,
            "sep"               : "|",
            "width"             : 80,
            "lines"             : 10,
            "fullscreen"        : False,
            "normal_window"     : False,
            "fixed_lines"       : 20,
            "markup_rows"       : True,
            "multi_select"      : True,
            "case_sensitive"    : False,
            "header-format"     : \
               "<b>{doc[title]}</b>\n"\
               "{doc[empty]}  <i>{doc[author]}</i>\n"
               "{doc[empty]}  <span foreground=\"red\">({doc[year]:->4})</span>"\
               "<span foreground=\"green\">{doc[tags]}</span>",
        },
    }



import re
import papis.downloaders.base


class Downloader(papis.downloaders.base.Downloader):
    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(self, url, name="get")

    @classmethod
    def match(cls, url):
        """
        >>> Downloader.match('http://wha2341!@#!@$%!@#file.pdf') is False
        False
        >>> Downloader.match('https://whateverpt?is?therefile.epub') is False
        False
        >>> not Downloader.match('http://whatever?path?is?therefile')
        True
        """
        endings = "pdf|djvu|epub|mobi|jpg|png|md"
        if re.match(r"^http.*\.(%s)$" % endings, url, re.IGNORECASE):
            return Downloader(url)
        else:
            return False

    def get_document_url(self):
        return self.get_url()



"""
ISBNPLUS Api Description
========================

Key             Description
---------------------------
q               Keywords, search for everything
p               Current page number, default is 1
a               Keywords, search for Author
c               Keywords, search for Catagory
s               Keywords, search for book series
t               Keywords, search for book title
order           published - return results by published years, newest first
                isbn - return results by isbn numbers, smaller number first
                any - return results by isbn numbers, smaller number first
                any - return results by relevance, most relevant results first
                default is published
app_id          Application ID, required, you can get your application ID after
                login
app_key         Application Key, required, you can get your application key
                after login

An example of sucessful returns:
================================

    <?xml version="1.0" encoding="UTF-8"?>
    <response status="ok">
    <page name="search">
    <count>71</count>
    <total>71</total>
    <pages>8</pages>
    <current_page>1</current_page>
    <results>
    <book>
    <ISBNPlus_id>LOC.V40.196960-1-3374190</ISBNPlus_id>
    <link>http://isbnplus.org/9780415887229</link>
    <isbn13>9780415887229</isbn13>
    <isbn10>0415887224</isbn10>
    <title>
    International News In The Digital Age: East-West Perceptions Of A New World Order
    </title>
    <author>Judith Clarke; Michael Bromley</author>
    <published_place>New York</published_place>
    <publisher>Routledge</publisher>
    <published_year>2012</published_year>
    <pages>234</pages>
    <language>English</language>
    <lccn>2011008858</lccn>
    <format/>
    <series>Routledge research in journalism ( Volume 4 )</series>
    <reisbn>0</reisbn>
    <category>
    Journalism > International Cooperation; Citizen Journalism > Political Aspects; Communication, International > Political Aspects; Online Journalism; Mass Media And International Relations; Foreign News > Political Aspects;
    </category>
    </book>
    ...
"""
import xml.dom.minidom
import urllib.request# urlopen, Request
import urllib.parse# import urlencode
import bs4
import papis.config
import logging

logger = logging.getLogger('isbn')

ISBNPLUS_KEY = "98a765346bc0ffee6ede527499b6a4ee"
ISBNPLUS_APPID = "4846a7d1"
ISBNPLUS_BASEURL = "https://api-2445581351187.apicast.io:443/"
ISBNPLUS_BASEURL = "https://api-2445581351187.apicast.io/"

def get_data(
    query="",
    page=1,
    author="",
    category="",
    series="",
    title="",
    order="isbn",
    app_id=ISBNPLUS_APPID,
    app_key=ISBNPLUS_KEY
    ):
    results = []
    dict_params = {
        "q": query,
        "p": page,
        "a": author,
        "c": category,
        "s": series,
        "t": title,
        "order": order,
        "app_id": app_id,
        "app_key": app_key
    }
    params = urllib.parse.urlencode(
        {x:dict_params[x] for x in dict_params if dict_params[x]}
    )
    req_url = ISBNPLUS_BASEURL + "search?" + params
    logger.debug("url = " + req_url)
    url = urllib.request.Request(
        req_url,
        headers={
            'User-Agent': papis.config.get('user-agent')
        }
    )
    xmldoc = urllib.request.urlopen(url).read()
    root = bs4.BeautifulSoup(xmldoc, 'html.parser')

    for book in root.find_all('book'):
        book_data = book_to_data(book)
        results.append(book_data)
    logger.debug('%s records retrieved' % len(results))
    return results


def book_to_data(booknode):
    """Convert book xml node into dictionary

    :booknode: Bs4 book node
    :returns: Dictionary containing its data

    """
    book = dict()
    keys_translate = [
        ("published_year", "year"), ("link", "url"),
    ]
    keys = [
        "title", "author", "language", "publisher", "pages", "isbn10",
        "isbn13", "link", "keywords", "published_year", "published_place",
        "series", "lccn", "ISBNPlus_id", "year"
    ]
    for key in keys:
        val_list = booknode.find_all(key)
        if len(val_list):
            book[key] = val_list[0].text
    for key_pair in keys_translate:
        if key_pair[0] in book.keys():
            book[key_pair[1]] = book[key_pair[0]]
    return book



from subprocess import call
import logging

logger = logging.getLogger("utils")
logger.debug("importing")

import os
import re
import string
import papis.api
import papis.config
import papis.commands
import papis.document
import papis.crossref
import papis.bibtex


def general_open(fileName, key, default_opener="xdg-open", wait=False):
    try:
        opener = papis.config.get(key)
    except KeyError:
        opener = default_opener
    if isinstance(fileName, list):
        fileName = papis.api.pick(fileName)
    if isinstance(opener, str):
        if wait:
            return os.system(" ".join([opener, fileName]))
        else:
            return call([opener, fileName])
    elif hasattr(opener, '__call__'):
        return opener(fileName)
    else:
        raise Warning("How should I use the opener %s?" % opener)


def get_regex_from_search(search):
    """Creates a default regex from a search string.

    :param search: A valid search string
    :type  search: str
    :returns: Regular expression
    :rtype: str
    """
    return r".*"+re.sub(r"\s+", ".*", search)


def format_doc(python_format, document):
    """Construct a string using a pythonic format string and a document.

    :param python_format: Python-like format string.
        (`see <
            https://docs.python.org/2/library/string.html#format-string-syntax
        >`_)
    :type  python_format: str
    :param document: Papis document
    :type  document: papis.document.Document
    :returns: Formated string
    :rtype: str
    """
    doc = papis.config.get("format-doc-name")
    return python_format.format(**{doc: document})


def match_document(document, search, match_format=None):
    """Main function to match document to a given search.

    :param document: Papis document
    :type  document: papis.document.Document
    :param search: A valid search string
    :type  search: str
    :param match_format: Python-like format string.
        (`see <
            https://docs.python.org/2/library/string.html#format-string-syntax
        >`_)
    :type  match_format: str
    :returns: Non false if matches, true-ish if it does match.
    """
    match_format = match_format or papis.config.get("match-format")
    match_string = format_doc(match_format, document)
    regex = get_regex_from_search(search)
    return re.match(regex, match_string, re.IGNORECASE)


def get_folders(folder):
    """This is the main indexing routine. It looks inside ``folder`` and crawls
    the whole directory structure in search for subfolders containing an info
    file.

    :param folder: Folder to look into.
    :type  folder: str
    :returns: List of folders containing an info file.
    :rtype: list
    """
    logger.debug("Indexing folders")
    folders = list()
    for root, dirnames, filenames in os.walk(folder):
        if os.path.exists(os.path.join(root, get_info_file_name())):
            folders.append(root)
    return folders


class DocMatcher(object):
    """This class implements the mini query language for papis.
    All its methods are static, it could be also implemented as a separate
    module.

    The static methods are to be used as follows:
    First the search string has to be set,
        DocMatcher.set_search(search_string)
    and then the parse method should be called in order to decypher the
    search_string,
        DocMatcher.parse()
    Now the DocMatcher is ready to match documents with the input query
    via the `return_if_match` method, which is used to parallelize the
    matching.
    """
    import papis.config
    search = ""
    parsed_search = None
    doc_format = '{' + papis.config.get('format-doc-name') + '[DOC_KEY]}'
    logger = logging.getLogger('DocMatcher')

    @classmethod
    def return_if_match(cls, doc):
        """Use the attribute `cls.parsed_search` to match the `doc` document
        to the previously parsed query.
        :param doc: Papis document to match against.
        :type  doc: papis.document.Document
        :returns: True if it matches, False if some query requirement does
            not match.
        """
        match = None
        for parsed in cls.parsed_search:
            if len(parsed) == 1:
                search = parsed[0]
                sformat = None
            elif len(parsed) == 3:
                search = parsed[2]
                sformat = cls.doc_format.replace('DOC_KEY',parsed[0])
            match = doc if papis.utils.match_document(
                doc, search, match_format=sformat) else None
            if not match:
                break
        return match

    @classmethod
    def set_search(cls, search):
        cls.search = search

    @classmethod
    def parse(cls, search=False):
        """Parse the main query text. This method will also set the
        class attribute `parsed_search` to the parsed query, and it will
        return it too.
        :param cls: The class object, since it is a static method
        :type  cls: object
        :param search: Search text string if a custom search string is to be
            used. False if the `cls.search` class attribute is to be used.
        :type  search: str
        :returns: Parsed query
        :rtype:  list
        >>> print(DocMatcher.parse('hello author = einstein'))
        [['hello'], ['author', '=', 'einstein']]
        >>> print(DocMatcher.parse(''))
        []
        >>> print(\
            DocMatcher.parse(\
                '"hello world whatever =" tags = \\\'hello ====\\\''))
        [['hello world whatever ='], ['tags', '=', 'hello ====']]
        >>> print(DocMatcher.parse('hello'))
        [['hello']]
        """
        import pyparsing
        cls.logger.debug('Parsing search')
        search = search or cls.search
        papis_alphas = pyparsing.printables.replace('=', '')
        papis_key = pyparsing.Word(pyparsing.alphanums)
        papis_value = pyparsing.QuotedString(
            quoteChar='"', escChar='\\', escQuote='\\'
        ) ^ pyparsing.QuotedString(
            quoteChar="'", escChar='\\', escQuote='\\'
        ) ^ papis_key
        equal = pyparsing.ZeroOrMore(" ") + \
                pyparsing.Literal('=')    + \
                pyparsing.ZeroOrMore(" ")

        papis_query = pyparsing.ZeroOrMore(
            pyparsing.Group(
                pyparsing.ZeroOrMore(
                    papis_key + equal
                ) + papis_value
            )
        )
        parsed = papis_query.parseString(search)
        cls.logger.debug('Parsed search = %s' % parsed)
        cls.parsed_search = parsed
        return cls.parsed_search



def filter_documents(documents, search=""):
    """Filter documents. It can be done in a multi core way.

    :param documents: List of papis documents.
    :type  documents: papis.documents.Document
    :param search: Valid papis search string.
    :type  search: str
    :returns: List of filtered documents
    :rtype:  list

    """
    logger = logging.getLogger('filter')
    papis.utils.DocMatcher.set_search(search)
    papis.utils.DocMatcher.parse()
    if search == "" or search == ".":
        return documents
    else:
        # Doing this multiprocessing in filtering does not seem
        # to help much, I don't know if it's because I'm doing something
        # wrong or it is really like this.
        import multiprocessing
        import time
        np = papis.api.get_arg("cores", multiprocessing.cpu_count())
        pool = multiprocessing.Pool(np)
        logger.debug(
            "Filtering docs (search %s) using %s cores" % (
                search,
                np
            )
        )
        logger.debug("pool started")
        begin_t = time.time()
        result = pool.map(
            papis.utils.DocMatcher.return_if_match, documents
        )
        pool.close()
        pool.join()
        logger.debug("pool done (%s ms)" % (1000*time.time()-1000*begin_t))
        return [d for d in result if d is not None]


def get_documents(directory, search=""):
    """Get documents from within a containing folder

    :param directory: Folder to look for documents.
    :type  directory: str
    :param search: Valid papis search
    :type  search: str
    :returns: List of document objects.
    :rtype: list
    """
    import papis.config
    directory = os.path.expanduser(directory)

    if papis.config.getboolean("use-cache"):
        import papis.cache
        folders = papis.cache.get_folders(directory)
    else:
        folders = get_folders()

    logger.debug("Creating document objects")
    documents = folders_to_documents(folders)
    logger.debug("Done")

    return filter_documents(documents, search)


def folders_to_documents(folders):
    """Turn folders into documents, this is done in a multiprocessing way, this
    step is quite critical for performance.

    :param folders: List of folder paths.
    :type  folders: list
    :returns: List of document objects.
    :rtype:  list
    """
    import multiprocessing
    import time
    logger = logging.getLogger("dir2doc")
    np = papis.api.get_arg("cores", multiprocessing.cpu_count())
    logger.debug("Running in %s cores" % np)
    pool = multiprocessing.Pool(np)
    logger.debug("pool started")
    begin_t = time.time()
    result = pool.map(papis.document.Document, folders)
    pool.close()
    pool.join()
    logger.debug("pool done (%s ms)" % (1000*time.time()-1000*begin_t))
    return result


def folder_is_git_repo(folder):
    """Check if folder is a git repository

    :folder: Folder to check
    :returns: Wether is git repo or not
    :rtype:  bool

    """
    import subprocess
    logger.debug("Check if %s is a git repo" % folder)
    try:
        subprocess.check_call(
            ' '.join(['git', '-C', folder, 'status']),
            stdout=None,
            shell=True
        )
        logger.debug("Detected git repo in %s" % folder)
        return True
    except:
        return False


def lib_is_git_repo(library):
    """Check if library is a git repository

    :library: Library to check
    :returns: Wether is git repo or not
    :rtype:  bool
    """
    config = papis.config.get_configuration()
    return folder_is_git_repo(config.get(library, "dir"))


def get_info_file_name():
    """Get the name of the general info file for any document

    :returns: Name of the file.
    :rtype: str
    """
    return papis.config.get("info-name")


def doi_to_data(doi):
    """Try to get from a DOI expression a dictionary with the document's data
    using the crossref module.

    :param doi: DOI expression.
    :type  doi: str
    :returns: Document's data
    :rtype: dict
    """
    return papis.crossref.doi_to_data(doi)


def yaml_to_data(yaml_path):
    """Convert a yaml file into a dictionary using the yaml module.

    :param yaml_path: Path to a yaml file
    :type  yaml_path: str
    :returns: Dictionary containing the info of the yaml file
    :rtype:  dict
    """
    import yaml
    return yaml.load(open(yaml_path))


def vcf_to_data(vcard_path):
    """Convert a vcf file into a dictionary using the vobject module.

    :param vcf_path: Path to a vcf file
    :type  vcf_path: str
    :returns: Dictionary containing the info of the vcf file
    :rtype:  dict
    """
    import vobject
    import yaml
    import papis.document.Document
    data = yaml.load(papis.document.Document.get_vcf_template())
    logger.debug("Reading in %s " % vcard_path)
    text = open(vcard_path).read()
    vcard = vobject.readOne(text)
    try:
        data["first_name"] = vcard.n.value.given
        logger.debug("First name = %s" % data["first_name"])
    except:
        data["first_name"] = None
    try:
        data["last_name"] = vcard.n.value.family
        logger.debug("Last name = %s" % data["last_name"])
    except:
        data["last_name"] = None
    try:
        if not isinstance(vcard.org.value[0], list):
            data["org"] = vcard.org.value
        else:
            data["org"] = vcard.org.value
        logger.debug("Org = %s" % data["org"])
    except:
        data["org"] = []
    for ctype in ["tel", "email"]:
        try:
            vcard_asset = getattr(vcard, ctype)
            logger.debug("Parsing %s" % ctype)
        except:
            pass
        else:
            try:
                param_type = getattr(vcard_asset, "type_param")
            except:
                param_type = "home"
            data[ctype][param_type.lower()] = getattr(vcard_asset, "value")
    logger.debug("Read in data = %s" % data)
    return data


def confirm(prompt, yes=True):
    """Confirm with user input

    :param prompt: Question or text that the user gets.
    :type  prompt: str
    :param yes: If yes should be the default.
    :type  yes: bool
    :returns: True if go ahead, False if stop
    :rtype:  bool

    """
    import prompt_toolkit
    result = prompt_toolkit.prompt(
        prompt + ' (%s): ' % ('Y/n' if yes else 'y/N')
    )
    if yes:
        return result not in ['N', 'n']
    else:
        return result not in ['Y', 'y']


def input(prompt, default=""):
    """Prompt user for input

    :param prompt: Question or text that the user gets.
    :type  prompt: str
    :param default: Default value to give if the user does not input anything
    :type  default: str
    :returns: User input or default
    :rtype:  bool

    """
    import prompt_toolkit
    result = prompt_toolkit.prompt(
        prompt + ' (%s): ' % (default)
    )
    return result if result else default


def clean_document_name(doc_path):
    """Get a file path and return the basename of the path cleaned.

    :param doc_path: Path of a document.
    :type  doc_path: str
    :returns: Basename of the path cleaned
    :rtype:  str

    >>> clean_document_name('{{] __ }}albert )(*& $ß $+_ einstein (*]')
    'albert-ss-_-einstein'
    >>> clean_document_name('/ashfd/df/  #$%@#$ }{_+"[ ]hello öworld--- .pdf')
    'hello-oworld----.pdf'
    """
    import unidecode
    base = os.path.basename(doc_path)
    logger.debug("Cleaning document name %s " % base)
    trans_dict = dict.fromkeys(
        string.punctuation.translate(
          str.maketrans(dict.fromkeys('.-_'))
        )
    )
    translation = str.maketrans(trans_dict)
    cleaned = base.translate(translation)
    cleaned = cleaned.strip(string.whitespace+string.punctuation)
    cleaned = cleaned.strip()
    cleaned = re.sub(r"\s+", "-", cleaned)
    cleaned = unidecode.unidecode(cleaned)
    return cleaned


def git_commit(path="", message=""):
    """Commits changes in the path with a message.
    If the path is not given, then the lib path is used.

    :param path: Folder where a git repo exists.
    :type  path: str
    :param message: Commit message
    :type  message: str
    :returns: None

    """
    logger.debug('Commiting...')
    path = path or os.path.expanduser(papis.config.get('dir'))
    message = '-m "%s"' % message if len(message) > 0 else ''
    cmd = ['git', '-C', path, 'commit', message]
    logger.debug(cmd)
    message = '-m "%s"' % message if len(message) > 0 else ''
    call(cmd)


def file_is(file_description, fmt):
    """Get if file stored in `file_path` is a `fmt` document.

    :file_path: Full path for a `fmt` file or a buffer containing `fmt` data.
    :returns: True if is `fmt` and False otherwise

    """
    import magic
    logger.debug("Checking filetype")
    if os.path.exists(file_description):
        result = re.match(
            r".*%s.*" % fmt, magic.from_file(file_description, mime=True)
        )
        if result:
            logger.debug(
                "File %s appears to be of type %s" % (file_description, fmt)
            )
    else:
        # Suppose is a buffer
        result = re.match(
            r".*%s.*" % fmt, magic.from_buffer(file_description, mime=True)
        )
        if result:
            logger.debug(
                "Buffer appears to be of type %s" % (fmt)
            )
    return True if result else False


def is_pdf(file_description):
    return file_is(file_description, 'pdf')


def is_djvu(file_description):
    return file_is(file_description, 'djvu')


def is_epub(file_description):
    return file_is(file_description, 'epub')


def is_mobi(file_description):
    return file_is(file_description, 'mobi')



"""
The command config is a useful command because it allows you to check
the configuration settings' values that your current `papis` session
is using.

For example let's say that you want to see which ``dir`` setting your
current library is using (i.e., the directory or the dir that appears
in the definition of the library in the configuration file), then you
would simply do:

.. code::

    papis config dir

If you wanted to see which ``dir`` the library ``books`` has, for example
then you would do

.. code::

    papis -l books config dir

This works as well for default settings, i.e., settings that you have not
customized, for example the setting ``match-format``, you would check
it with

.. code::

    papis config match-format
    > {doc[tags]}{doc.subfolder}{doc[title]}{doc[author]}{doc[year]}

You can find a list of all available settings in the configuration section.


"""
import sys
import os
import re
import configparser
import papis.commands


class Command(papis.commands.Command):
    def init(self):

        self.parser = self.get_subparsers().add_parser(
            "config",
            help="Print configuration values"
        )

        self.parser.add_argument(
            "option",
            help="Variable to print",
            default="",
            action="store"
        )

    def main(self):
        option = self.args.option.split(".")
        self.logger.debug(option)
        if len(option) == 1:
            key = option[0]
            section = None
        elif len(option) == 2:
            section = option[0]
            key = option[1]
        self.logger.debug("key = %s, sec = %s" % (key, section))
        val = papis.config.get(key, section=section)
        print(val)



import papis
import sys
import os
import re
import papis.api
import papis.utils
import subprocess


class Command(papis.commands.Command):
    def init(self):

        self.parser = self.get_subparsers().add_parser(
            "mv",
            help="Move entry"
        )

        self.add_search_argument()
        self.add_git_argument()

    def get_dirs(self, main):
        directories = []
        p = ""
        for root, dirs, files in os.walk(main):
            for di in dirs:
                p = os.path.join(root, di, papis.utils.get_info_file_name())
                if not os.path.exists(p) \
                   and not re.match(r".*[.]git.*", os.path.join(root, di)):
                    directories.append(di)
        self.logger.debug(directories)
        return directories

    def main(self):

        # Leave this imports here for performance
        import prompt_toolkit
        import prompt_toolkit.contrib.completers

        documents = papis.api.get_documents_in_lib(
            self.get_args().lib,
            self.get_args().search
        )

        document = self.pick(documents)
        if not document: return 0

        lib_dir = os.path.expanduser(papis.config.get('dir'))
        folder = document.get_main_folder()

        directories = self.get_dirs(lib_dir)

        completer = prompt_toolkit.contrib.completers.WordCompleter(
            directories
        )

        try:
            new_folder = os.path.join(
                lib_dir,
                prompt_toolkit.prompt(
                    "Enter directory: (Tab completion enabled)\n"
                    ">  ",
                    completer=completer
                )
            )
        except:
            return 0

        self.logger.info(new_folder)

        if not os.path.exists(new_folder):
            self.logger.info("Creating path %s" % new_folder)
            os.makedirs(new_folder)

        mvtool = papis.config.get("mvtool")

        cmd = (['git', '-C', folder] if self.args.git else []) + \
            ['mv', folder, new_folder]
        self.logger.debug(cmd)
        subprocess.call(cmd)
        papis.api.clear_lib_cache()



"""
This is the command that can be used to synchronize libraries.
You can customize the sync command using the configuration setting
`sync-command`.

Examples of `sync-command` setting usage are:

    - For git compatibility, use for instance

    .. code::

        sync-command = git -C {lib[dir]} pull origin master

    - For rsync compatibility, use for instance

    .. code::

        sync-command = rsync -r some/remote/host {lib[dir]}

"""
import os
import string
import papis.commands
import papis.config
import papis.api


class Command(papis.commands.Command):

    def init(self):
        self.parser = self.get_subparsers().add_parser(
            "sync",
            help="Sync a library using the sync command"
        )

    def main(self):
        sync_command = os.path.expanduser(
            papis.config.get("sync-command")
        )
        command = sync_command.format(
            lib=self.get_config()[self.get_args().lib]
        )
        print(command)
        os.system(command)
        papis.api.clear_lib_cache()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Heavily adapted for papis from
#   author: François-Xavier Coudert
#   license: MIT License
#   src: https://github.com/fxcoudert/tools/blob/master/doi2bib
#
from __future__ import unicode_literals
import logging
logger = logging.getLogger("crossref")
logger.debug("importing")

import sys
import unicodedata
import re
import papis.config

# CrossRef queries
#
# CrossRef documentation comes from here:
# http://labs.crossref.org/site/quick_and_dirty_api_guide.html
#
# You need a CrossRef API key. 
#
CROSSREF_KEY = "fx.coudert@chimie-paristech.fr"
CROSSREF_KEY = "a.gallo@fkf.mpg.de"
#
# Using Google allows one to find other API keys:
# zter:zter321
# ourl_rdmpage:peacrab
# egon@spenglr.com
# s_allannz@yahoo.com
# dollar10boy@hotmail.com

# LaTeX accents replacement
latex_accents = {
  "à": "\\`a",  # Grave accent
  "è": "\\`e",
  "ì": "\\`{\\i}",
  "ò": "\\`o",
  "ù": "\\`u",
  "ỳ": "\\`y",
  "À": "\\`A",
  "È": "\\`E",
  "Ì": "\\`{\\I}",
  "Ò": "\\`O",
  "Ù": "\\`U",
  "Ỳ": "\\`Y",
  "á": "\\'a",  # Acute accent
  "ć": "\\'c",
  "é": "\\'e",
  "í": "\\'{\\i}",
  "ó": "\\'o",
  "ú": "\\'u",
  "ý": "\\'y",
  "Á": "\\'A",
  "É": "\\'E",
  "Í": "\\'{\\I}",
  "Ó": "\\'O",
  "Ú": "\\'U",
  "Ý": "\\'Y",
  "â": "\\^a",  # Circumflex
  "ê": "\\^e",
  "î": "\\^{\\i}",
  "ô": "\\^o",
  "û": "\\^u",
  "ŷ": "\\^y",
  "Â": "\\^A",
  "Ê": "\\^E",
  "Î": "\\^{\\I}",
  "Ô": "\\^O",
  "Û": "\\^U",
  "Ŷ": "\\^Y",
  "ä": "\\\"a",  # Umlaut or dieresis
  "ë": "\\\"e",
  "ï": "\\\"{\\i}",
  "ö": "\\\"o",
  "ü": "\\\"u",
  "ÿ": "\\\"y",
  "Ä": "\\\"A",
  "Ë": "\\\"E",
  "Ï": "\\\"{\\I}",
  "Ö": "\\\"O",
  "Ü": "\\\"U",
  "Ÿ": "\\\"Y",
  "ã": "\\~{a}",  # Tilde
  "ñ": "\\~{n}",
  "ă": "\\u{a}",  # Breve
  "ĕ": "\\u{e}",
  "ŏ": "\\u{o}",
  "š": "\\v{s}",  # Caron
  "č": "\\v{c}",
  "ç": "\\c{c}",  # Cedilla
  "Ç": "\\c{C}",
  "œ": "{\\oe}",  # Ligatures
  "Œ": "{\\OE}",
  "æ": "{\\ae}",
  "Æ": "{\\AE}",
  "å": "{\\aa}",
  "Å": "{\\AA}",
  "–": "--",  # Dashes
  "—": "---",
  "−": "--",
  "ø": "{\\o}",  # Misc latin-1 letters
  "Ø": "{\\O}",
  "ß": "{\\ss}",
  "¡": "{!`}",
  "¿": "{?`}",
  "\\": "\\\\",  # Characters that should be quoted
  "~": "\\~",
  "&": "\\&",
  "$": "\\$",
  "{": "\\{",
  "}": "\\}",
  "%": "\\%",
  "#": "\\#",
  "_": "\\_",
  "≥": "$\\ge$",  # Math operators
  "≤": "$\\le$",
  "≠": "$\\neq$",
  "©": "\copyright", # Misc
  "ı": "{\\i}",
  "α": "$\\alpha$",
  "β": "$\\beta$",
  "γ": "$\\gamma$",
  "δ": "$\\delta$",
  "ε": "$\\epsilon$",
  "η": "$\\eta$",
  "θ": "$\\theta$",
  "λ": "$\\lambda$",
  "µ": "$\\mu$",
  "ν": "$\\nu$",
  "π": "$\\pi$",
  "σ": "$\\sigma$",
  "τ": "$\\tau$",
  "φ": "$\\phi$",
  "χ": "$\\chi$",
  "ψ": "$\\psi$",
  "ω": "$\\omega$",
  "°": "$\\deg$",
  "‘": "`",  # Quotes
  "’": "'",
  "′": "$^\\prime$",
  "“": "``",
  "”": "''",
  "‚": ",",
  "„": ",,",
  "\xa0": " ",     # Unprintable characters
}


def replace_latex_accents(string):
    s = unicodedata.normalize('NFC', string)
    return "".join([latex_accents[c] if c in latex_accents else c for c in s])


def validate_doi(doi):
    """We check that the DOI can be resolved by official means.  If so, we
    return the resolved URL, otherwise, we return None (which means the DOI is
    invalid).

    :param doi: Doi identificator
    :type  doi: str
    """
    from urllib.request import urlopen, Request
    handle_url = "http://dx.doi.org/" + doi
    logger.debug('handle url %s' % handle_url)
    try:
      handle = urlopen(handle_url)
    except:
      return None

    resolvedURL = handle.geturl()
    logger.debug('resolved url %s' % resolvedURL)
    if resolvedURL[0:18] == "http://dx.doi.org/":
      return None
    else:
      return resolvedURL


def get_citation_info_from_results(container):
    """This function retrieves the citations from the container's answer

    :param container: xml information
    :returns: Dictionary with information to be added
    :rtype:  dict

    """
    citations_info = dict(citations=[])
    logger.debug("Getting citations..")

    for node in container.getElementsByTagName("citation"):
        citation = dict()
        # Some documents do not have a doi:
        doi_node = node.getElementsByTagName('doi')
        if len(doi_node) == 0:
            continue
        doi = doi_node[0].firstChild.data
        citation[ 'doi' ] = doi
        citations_info['citations'].append(citation)

    return citations_info


def get_author_info_from_results(container):
    """This function retrieves the authors from the container answer

    :param container: xml information
    :returns: Dictionary with information to be added
    :rtype:  dict

    """
    logger.debug("Getting authors..")
    authors_info = dict(author_list=[], author=None)

    for node in container.getElementsByTagName("person_name"):
        author = dict()
        surname = node.getElementsByTagName('surname')[0].firstChild.data
        given_name = node.getElementsByTagName('given_name')[0].firstChild.data
        author[ 'surname' ] = surname
        author[ 'given_name' ] = given_name
        authors_info['author_list'].append(author)

    authors_info['author'] = ', '.join([
        "{au[given_name]} {au[surname]}".format(au=author)
        for author in authors_info['author_list']
    ])

    return authors_info


def get_cross_ref(doi):
    """Get the XML from CrossRef
    """
    global CROSSREF_KEY
    global logger
    import xml.dom.minidom
    from urllib.parse import urlencode
    from urllib.request import urlopen, Request
    params = urlencode({
        "id": "doi:" + doi,
        "noredirect": "true",
        "pid": CROSSREF_KEY,
        "format": "unixref"
    })
    req_url = "http://www.crossref.org/openurl/?" + params
    url = Request(
        req_url,
        headers={
            'User-Agent': papis.config.get('user-agent')
        }
    )
    doc = urlopen(url).read()
    logger.debug("Request url: %s" % req_url)

    # Parse it
    doc = xml.dom.minidom.parseString(doc)
    records = doc.getElementsByTagName("journal")

    # No results. Is it a valid DOI?
    if len(records) == 0:
        res = validate_doi(doi)
        if res is None:
            raise Exception("Invalid DOI")
        else:
            raise Exception("Can't locate metadata")

    if (len(records) != 1):
        raise Exception("CrossRef returned more than one record")

    record = records[0]

    # helper functions
    def find_item_named(container, name):
        obj_list = container.getElementsByTagName(name)
        if (len(obj_list) == 0):
            return None
        else:
            return obj_list[0]

    def data(node):
        if node is None:
            return None
        else:
            return node.firstChild.data

    res = dict()

    # JOURNAL INFO
    journal = find_item_named(record, "journal_metadata")
    if journal:
        res["full_journal_title"] = data(find_item_named(journal, "full_title"))
        res["abbrev_journal_title"] = data(find_item_named(journal, "abbrev_title"))

    # VOLUME INFO
    issue = find_item_named(record, "journal_issue")
    res["issue"] = data(find_item_named(issue, "issue"))
    res["volume"] = data(find_item_named(issue, "volume"))
    res["year"] = data(find_item_named(issue, "year"))

    # URLS INFO
    doi_resources = record\
        .getElementsByTagName('doi_data')[0]\
        .getElementsByTagName('item')
    for resource in doi_resources:
        if resource.hasAttribute('crawler'):
            key = papis.config.get('doc-url-key-name')
        else:
            key = 'url'
        if key:
            res[key] = resource.getElementsByTagName('resource')[0]\
                    .firstChild.data
        key = False

    # OTHER INFO
    other = find_item_named(record, "journal_article")
    res["title"] = data(find_item_named(other, "title"))
    res["first_page"] = data(find_item_named(other, "first_page"))
    res["last_page"] = data(find_item_named(other, "last_page"))
    if res["first_page"] is not None and res["last_page"] is not None:
        res['pages'] = res["first_page"] + "--" + res["last_page"]
    else:
        del res['first_page']
        del res['last_page']
    res["doi"] = data(find_item_named(other, "doi"))
    if res["year"] is None:
        res["year"] = data(find_item_named(other, "year"))

    # AUTHOR INFO
    res.update(get_author_info_from_results(record))

    # CITATION INFO
    res.update(get_citation_info_from_results(record))

    # REFERENCE BUILDING
    res["ref"] = doi

    # Journal checking
    # If the key journal does not exist check for abbrev_journal_title
    # and full_journal_title and set it then to one of them
    if 'journal' not in res.keys():
        for key in ['abbrev_journal_title', 'full_journal_title']:
            if key in res.keys():
                res['journal'] = res[key]

    return res


def get_clean_doi(doi):
    """Check if doi is actually a url and in that case just get
    the exact doi.

    :doi: String containing a doi
    :returns: The pure doi

    >>> get_clean_doi('http://dx.doi.org/10.1063/1.881498')
    '10.1063/1.881498'
    >>> get_clean_doi('10.1063/1.881498')
    '10.1063/1.881498'
    >>> get_clean_doi('http://physicstoday.scitation.org/doi/10.1063/1.881498')
    '10.1063/1.881498'
    >>> get_clean_doi('http://physicstoday.scitation.org/doi/10.1063/1.881498?asdfwer')
    '10.1063/1.881498'
    """
    mdoi = re.match(r'(.*doi(.org)?/)?(.*/[^?&%^$]*).*', doi)
    if mdoi:
        return mdoi.group(3)
    else:
        return None

def doi_to_data(doi):
    """Search through crossref and get a dictionary containing the data

    :param doi: Doi identificator
    :type  doi: str
    :returns: Dictionary containing the data

    """
    global logger
    doi = get_clean_doi(doi)
    try:
        data = get_cross_ref(doi)
    except Exception as e:
        logger.error(
            "Couldn't resolve DOI '" + doi + "' through CrossRef: " + str(e) + "\n"
        )
        return dict()
    else:
        return data



import logging
import os
import papis.utils
import papis.config


logger = logging.getLogger("cache")


def get_folder():
    """Get folder where the cache files are stored, it retrieves the
    ``cache-dir`` configuration setting. It is ``XDG`` standard compatible.
    """
    return os.path.expanduser(papis.config.get('cache-dir'))


def get(path):
    """Get contents stored in a cache file ``path`` in pickle binary format.

    :param path: Path to the cache file.
    :type  path: str
    :returns: Content of the cache file.
    :rtype: object
    """
    import pickle
    logger.debug("Getting cache %s " % path)
    return pickle.load(open(path, "rb"))


def create(obj, path):
    """Create a cache file in ``path`` with obj as its content using pickle
    binary format.

    :param obj: Any seriazable object.
    :type  obj: object
    :param path: Path to the cache file.
    :type  path: str
    :returns: Nothing
    :rtype: None
    """
    import pickle
    logger.debug("Saving in cache %s " % path)
    pickle.dump(obj, open(path, "wb+"))


def get_name(directory):
    """Create a cache file name out of the path of a given directory.

    :param directory: Folder name to be used as a seed for the cache name.
    :type  directory: str
    :returns: Name for the cache file.
    :rtype:  str
    """
    import hashlib
    return hashlib\
           .md5(directory.encode())\
           .hexdigest()+"-"+os.path.basename(directory)


def clear(directory):
    """Clear cache associated with a directory

    :param directory: Folder name that was used as a seed for the cache name.
    :type  directory: str
    :returns: Nothing
    :rtype: None
    """
    directory = os.path.expanduser(directory)
    cache_name = get_name(directory)
    cache_path = os.path.join(get_folder(), cache_name)
    if os.path.exists(cache_path):
        logger.debug("Clearing cache %s " % cache_path)
        os.remove(cache_path)


def clear_lib_cache(lib=None):
    """Clear cache associated with a library. If no library is given
    then the current library is used.

    :param lib: Library name.
    :type  lib: str
    """
    directory = papis.config.get("dir", section=lib)
    clear(directory)


def get_folders(directory):
    """Get folders from within a containing folder from cache

    :param directory: Folder to look for documents.
    :type  directory: str
    :param search: Valid papis search
    :type  search: str
    :returns: List of document objects.
    :rtype: list
    """
    cache = get_folder()
    cache_name = get_name(directory)
    cache_path = os.path.join(cache, cache_name)
    folders = []
    logger.debug("Getting documents from dir %s" % directory)
    logger.debug("Cache path = %s" % cache_path)
    if not os.path.exists(cache):
        logger.debug("Creating cache dir %s " % cache)
        os.makedirs(cache)
    if os.path.exists(cache_path):
        logger.debug("Loading folders from cache")
        folders = get(cache_path)
    else:
        folders = papis.utils.get_folders(directory)
        create(folders, cache_path)
    return folders
